import numpy as np
import os
import h5py
from scipy.ndimage import gaussian_filter1d
from joblib import Parallel, delayed
from tqdm import tqdm
from .utils import spikeLocation, computeAutoCorr
import matplotlib
matplotlib.use('Agg')  # Use a non-interactive backend for matplotlib
import matplotlib.pyplot as plt

def preprocess(user_settings):
    """Preprocess the data and save the features.
    Compute the ISI, autocorrelogram, and location of each unit and save the features to the output folder.

    Arguments:
        - user_settings (dict): User settings

    Outputs:
        - locations.npy: The location of each unit in the 3D space
        - amplitude.npy: The amplitude of each unit
        - peak_channels.npy: The peak channel of each unit
        - auto_corr.npy: The autocorrelogram of each unit
        - isi.npy: The ISI of each unit
        - peth.npy: The peri-event time histogram of each unit
        - waveforms_centered.npy: The centered waveforms of each unit
    
    """

    # load the data
    print('Loading data...')
    data_folder = user_settings["path_to_data"]
    waveform_all = np.load(os.path.join(data_folder, 'waveform_all.npy'))
    channel_locations = np.load(os.path.join(data_folder, 'channel_locations.npy'))
    sessions = np.load(os.path.join(data_folder, 'session_index.npy'))

    n_unit = waveform_all.shape[0]
    spike_times_all = [
        np.load(os.path.join(data_folder, 'spike_times/', 'Unit'+str(k)+'.npy')) for k in range(n_unit)]

    peth = None
    if os.path.isfile(os.path.join(data_folder, 'peth.npy')):
        peth = np.load(os.path.join(data_folder, 'peth.npy'))

    # make a folder to store the data
    output_folder = user_settings["output_folder"]
    if not os.path.isdir(output_folder):
        os.makedirs(output_folder)
    print(f'The output will be saved to {output_folder}!')

    # make a folder to store the figures
    figures_folder = os.path.join(output_folder, 'Figures')
    if not os.path.isdir(figures_folder):
        os.makedirs(figures_folder)

    # validate the data
    n_session = np.max(sessions)
    if n_session != len(np.unique(sessions)):
        raise ValueError('SessionIndex should start from 1 and be coninuous without any gaps!')

    print(n_session, 'sessions found!')

    # preprocessing data
    def process_spike_info(waveform, spike_times):
        # compute the location of each unit
        x, y, z, amp = spikeLocation(waveform, channel_locations,
                                    user_settings['spikeLocation']['n_nearest_channels'],
                                    user_settings['spikeLocation']['location_algorithm'])
        
        # compute the peak channel
        peaks_to_trough = np.max(waveform, axis=1) - np.min(waveform, axis=1)
        channel = np.argmax(peaks_to_trough)

        # centering the waveforms
        if user_settings['centering_waveforms']:
            n_channel = np.shape(waveform)[0]
            n_sample = np.shape(waveform)[1]
            center_position = np.ceil(n_sample/2) - 1

            # get the trough positions
            waveform_peak = np.squeeze(waveform[channel, :])
            idx_min = np.argmin(waveform_peak)

            # compute the delay
            delay = np.int64(center_position - idx_min)
            # if delay != 0:
            #     print(f'Correcting unit with delay = {delay} ...')

            # filling the borders with nearest values
            if delay == 0:
                waveforms_centered = waveform
            elif delay > 0:
                waveforms_centered = np.concatenate((waveform[:,0][:, np.newaxis] * np.ones((n_channel, delay)), np.squeeze(waveform[:,:n_sample-delay])), axis=1)
            else:
                waveforms_centered = np.concatenate((waveform[:,-delay:], waveform[:,-1][:, np.newaxis] * np.ones((n_channel, -delay))), axis=1)

            assert(np.shape(waveforms_centered) == np.shape(waveform))
        
        spike_times = spike_times - np.min(spike_times)

        # compute the autocorrelogram feauture
        auto_corr = None
        if "AutoCorr" in user_settings['motionEstimation']['features'] or \
                "AutoCorr" in user_settings['clustering']['features']:
            window = user_settings['autoCorr']['window']  # ms
            binwidth = user_settings['autoCorr']['binwidth']  # ms
            
            auto_corr, lag = computeAutoCorr(spike_times, window, binwidth)

            auto_corr[lag<0] = gaussian_filter1d(auto_corr[lag<0], user_settings['autoCorr']['gaussian_sigma'])
            auto_corr[lag>0] = gaussian_filter1d(auto_corr[lag>0], user_settings['autoCorr']['gaussian_sigma'])
            auto_corr = auto_corr / np.max(auto_corr)

        # compute the ISI feature
        isi_out = None
        if "ISI" in user_settings['motionEstimation']['features'] or \
                "ISI" in user_settings['clustering']['features']:
            isi = np.diff(spike_times)
            isi_hist = np.histogram(isi, bins=np.arange(0, user_settings['ISI']['window'], user_settings['ISI']['binwidth']))[0]
            isi_freq = isi_hist / np.sum(isi_hist)
            isi_out = gaussian_filter1d(isi_freq, user_settings['ISI']['gaussian_sigma'])
        
        return (x,y,z,amp,channel,auto_corr,isi_out,waveforms_centered)

    # print('Start preprocessing spikeInfo!')
    out = Parallel(n_jobs=user_settings["n_jobs"])(
        delayed(process_spike_info)(waveform_all[k], spike_times_all[k]) for k in tqdm(range(n_unit)))

    locations = np.zeros((n_unit, 3), dtype=np.float64)
    amp = np.zeros(n_unit, dtype=np.float64)
    channel = np.zeros(n_unit, dtype=np.int64)
    auto_corr = np.zeros((n_unit, np.shape(out[0][5])[0]), dtype=np.float64)
    isi = np.zeros((n_unit, int(user_settings['ISI']['window']/user_settings['ISI']['binwidth'])), dtype=np.float64)
    waveforms_centered = np.zeros((n_unit, waveform_all.shape[1], waveform_all.shape[2]), dtype=np.float64)

    for k in range(n_unit):
        locations[k, :] = out[k][0:3]
        amp[k] = out[k][3]
        channel[k] = out[k][4]
        auto_corr[k, :] = out[k][5]
        isi[k, :] = out[k][6]
        waveforms_centered[k, :, :] = out[k][7]

    # Save the preprocessed data
    print(f'Saving to {output_folder}...')

    np.save(os.path.join(output_folder, 'locations.npy'), locations)
    np.save(os.path.join(output_folder, 'amplitude.npy'), amp)
    np.save(os.path.join(output_folder, 'peak_channels.npy'), channel)
    np.save(os.path.join(output_folder, 'auto_corr.npy'), auto_corr)
    np.save(os.path.join(output_folder, 'isi.npy'), isi)
    np.save(os.path.join(output_folder, 'peth.npy'), peth)
    np.save(os.path.join(output_folder, 'waveforms_centered.npy'), waveforms_centered)
    print('Done!')

    # plot the data
    plt.figure(figsize=(10, 5))
    
    # plot the number of units in each session
    plt.subplot(121)
    n_unit_session = [np.sum(sessions==k) for k in range(1, n_session+1)]
    plt.plot(range(1, n_session+1), n_unit_session, 'k.-')
    plt.xlabel('Sessions')
    plt.ylabel('Number of units')
    plt.xlim([0.5, n_session+0.5])

    # plot the locations of the units in each session
    plt.subplot(122)
    x_plot = []
    y_plot = []
    x_scale = 0.8
    for k in range(n_unit):
        x_plot.append(sessions[k] + 1 + (np.random.rand()-0.5)*x_scale)
        y_plot.append(locations[k,1])
    
    plt.plot(x_plot, y_plot, '.', markersize=1, color='k', alpha=0.5)
    plt.xlabel('Sessions')
    plt.ylabel('Y location (μm)')
    plt.title('Unit locations')
    plt.xlim([0.5, n_session+0.5])

    plt.savefig(os.path.join(figures_folder, 'unitLocations.png'), dpi=300)
    plt.close()

def spikeInfo2npy(user_settings):
    """Convert the spikeInfo.mat file from MATLAB to numpy arrays that can be used in pyKilomatch.

    Arguments:
        - user_settings (dict): User settings

    Outputs:
        - waveform_all.npy: The waveform of each unit
        - session_index.npy: The session index of each unit
        - channel_locations.npy: The location of each channel
        - spike_times/: A folder that contains the spike times of each unit
        - spike_times/UnitA.npy: The spike times of unit A
        - peth.npy: The peri-event time histogram of each unit

    """

    # load the data
    data_folder = user_settings["path_to_data"]
    print('Loading', os.path.join(data_folder, 'spikeInfo.mat'), '...')
    spikeInfo = h5py.File(os.path.join(data_folder, 'spikeInfo.mat'))

    # make a folder to store the data
    data_folder = user_settings["path_to_data"]

    # Preprocess the spikeInfo
    n_unit = len(spikeInfo['spikeInfo']['RatName'])
    keys = spikeInfo['spikeInfo'].keys()

    Kcoords = np.array(spikeInfo[spikeInfo['spikeInfo']['Kcoords'][0][0]][0])
    Xcoords = np.array(spikeInfo[spikeInfo['spikeInfo']['Xcoords'][0][0]][0])
    Ycoords = np.array(spikeInfo[spikeInfo['spikeInfo']['Ycoords'][0][0]][0])
    SpikeTimes = [
        np.squeeze(
            np.array(spikeInfo[spikeInfo['spikeInfo']['SpikeTimes'][k][0]])) for k in range(n_unit)]
    SessionIndex = np.array([np.array(spikeInfo[spikeInfo['spikeInfo']['SessionIndex'][k][0]][0][0]) for k in range(n_unit)],
        dtype=np.int64)
    Waveform = [np.transpose(np.array(
        spikeInfo[spikeInfo['spikeInfo']['Waveform'][k][0]])) for k in range(n_unit)]

    PETH = None
    if "PETH" in keys:
        PETH = [np.squeeze(
            np.array(
                spikeInfo[spikeInfo['spikeInfo']['PETH'][k][0]])) for k in range(n_unit)]

    # validate the data
    n_session = np.max(SessionIndex)
    if n_session != len(np.unique(SessionIndex)):
        raise ValueError('SessionIndex should start from 1 and be coninuous without any gaps!')

    print(n_session, 'sessions found!')

    locations = np.zeros((n_unit, 3), dtype=np.float64)
    amp = np.zeros(n_unit, dtype=np.float64)
    channel = np.zeros(n_unit, dtype=np.int64)
    auto_corr = np.zeros((n_unit, user_settings['autoCorr']['window']), dtype=np.float64)
    isi = np.zeros((n_unit, int(user_settings['ISI']['window']/user_settings['ISI']['binwidth'])), dtype=np.float64)
    waveform_all = np.array(Waveform)
    channel_locations = np.column_stack((Xcoords, Ycoords))

    if PETH is not None:
        peth = np.zeros((n_unit, len(PETH[0])), dtype=np.float64)
        for k in range(n_unit):
            peth[k, :] = PETH[k]
    else:
        peth = None

    # Save the preprocessed data
    print(f'Saving to {data_folder}...')

    np.save(os.path.join(data_folder, 'waveform_all.npy'), waveform_all)
    np.save(os.path.join(data_folder, 'session_index.npy'), SessionIndex)
    np.save(os.path.join(data_folder, 'channel_locations.npy'), channel_locations)

    if not os.path.isdir(os.path.join(data_folder, 'spike_times/')):
        os.makedirs(os.path.join(data_folder, 'spike_times/'))

    for k in range(n_unit):
        np.save(os.path.join(data_folder,'spike_times/', 'Unit'+str(k)+'.npy'), SpikeTimes[k])

    if PETH is not None:
        np.save(os.path.join(data_folder, 'peth.npy'), peth)

    print('Done!')


import numpy as np
import os
from tqdm import tqdm
import hdbscan
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.neighbors import NearestNeighbors
from scipy.spatial.distance import squareform
from scipy.cluster.hierarchy import optimal_leaf_ordering, leaves_list
import matplotlib
matplotlib.use('Agg')  # Use a non-interactive backend for matplotlib
import matplotlib.pyplot as plt
from .utils import corrcoef2, Motion

def finalClustering(user_settings):
    """Final clustering of the units based on the similarity metrics using HDBSCAN and LDA.

    Arguments:
        - user_settings (dict): User settings
 
    """
    output_folder = user_settings["output_folder"]

    similarity_names = user_settings['clustering']['features']
    waveforms_corrected = np.load(os.path.join(output_folder, 'waveforms_corrected.npy'))
    motion = Motion.load(output_folder)

    iterativeClustering(user_settings, similarity_names, waveforms_corrected, motion)

def getNearbyPairs(max_distance, sessions, locations, motion=None):
    """Get the pairs of units that are within the max_distance.

    Arguments:
        - max_distance (float): The maximum distance between the units.
        - sessions (ndarray): The session index of the units.
        - locations (ndarray): The locations of the units.
        - motion (ndarray): The motion of the probe.

    Outputs:
        - idx_unit_pairs (ndarray): The pairs of units that are within the max_distance.
        - session_pairs (ndarray): The session index of the pairs of units.
    """
    n_unit = locations.shape[0]
    corrected_locations = np.zeros(n_unit)
    for k in range(n_unit):
        if motion is None:
            corrected_locations[k] = locations[k,1]
        else:
            corrected_locations[k] = locations[k,1] - motion.get_motion(sessions[k], locations[k,1])

    y_distance_matrix = np.abs(corrected_locations[:,np.newaxis] - corrected_locations[np.newaxis,:])
    idx_col = np.floor(np.arange(y_distance_matrix.size) / y_distance_matrix.shape[0]).astype(int)
    idx_row = np.mod(np.arange(y_distance_matrix.size), y_distance_matrix.shape[0]).astype(int)
    idx_good = np.where((y_distance_matrix.ravel() <= max_distance) & (idx_col > idx_row))[0]
    idx_unit_pairs = np.column_stack((idx_row[idx_good], idx_col[idx_good]))

    session_pairs = np.column_stack((sessions[idx_unit_pairs[:,0]], sessions[idx_unit_pairs[:,1]]))

    return idx_unit_pairs, session_pairs

def computeWaveformSimilarityMatrix(user_settings, waveforms, channel_locations):
    # Get waveform features
    n_nearest_channels = user_settings['waveformCorrection']['n_nearest_channels']
    n_unit = waveforms.shape[0]

    if len(waveforms.shape) > 3:
        n_templates = waveforms.shape[3]
    else:
        n_templates = 1
        waveforms = np.expand_dims(waveforms, axis=3)
    
    # find k-nearest neighbors for each channel
    nn = NearestNeighbors(n_neighbors=n_nearest_channels).fit(channel_locations)
    _, idx_nearest = nn.kneighbors(channel_locations)
    idx_nearest_sorted = np.sort(idx_nearest, axis=1)
    idx_nearest_unique, idx_groups = np.unique(idx_nearest_sorted, axis=0, return_inverse=True)

    # Compute the similarity matrix
    waveform_similarity_matrix = np.zeros((n_unit, n_unit, n_templates))

    for i_template in range(n_templates):
        waveform_similarity_matrix_this = np.zeros((n_unit, n_unit))
        ptt = np.squeeze(np.max(waveforms[:,:,:,i_template], axis=2) - np.min(waveforms[:,:,:,i_template], axis=2))
        ch = np.argmax(ptt, axis=1)

        for k in tqdm(range(idx_nearest_unique.shape[0]), desc='Computing waveform similarity'):
            idx_included = np.where(idx_groups == k)[0]
            idx_units = np.where(np.isin(ch, idx_included))[0]

            if len(idx_units) == 0:
                continue
            
            # The slice should be done twice to avoid indexing issues (different with MATLAB)
            waveform_this = waveforms[:,:,:,i_template]
            waveform_this = np.reshape(waveforms[:,idx_nearest_unique[k,:],:], (n_unit, -1))

            temp = corrcoef2(waveform_this[idx_units,:].T, waveform_this.T)
            temp[np.isnan(temp)] = 0
            temp = np.atanh(temp)
            
            waveform_similarity_matrix_this[idx_units,:] = temp
        
        waveform_similarity_matrix_this = np.max(np.stack((waveform_similarity_matrix_this, waveform_similarity_matrix_this.T), axis=2), axis=2)  
        waveform_similarity_matrix[:,:,i_template] = waveform_similarity_matrix_this

    waveform_similarity_matrix = np.max(waveform_similarity_matrix, axis=2)
    np.fill_diagonal(waveform_similarity_matrix, np.inf)

    return waveform_similarity_matrix

def computeAllSimilarityMatrix(user_settings, waveforms, feature_names):
    """Compute the similarity matrix of the units based on the similarity metrics.

    Arguments:
        - user_settings (dict): User settings
        - waveforms (ndarray): The waveforms of the units (n_units, n_channels, n_samples)
        - feature_names (list): The names of the features to be computed. The options are 'Waveform', 'ISI', 'AutoCorr', and 'PETH'.

    Outputs:
        - similarity_matrix_all (ndarray): The similarity matrix of the units (n_units, n_units, n_features)
        - feature_names_all (list): The names of the features computed.
        - waveform_similarity_matrix.npy: The waveform similarity matrix of the units (n_units, n_units)
        - ISI_similarity_matrix.npy: The ISI similarity matrix of the units (n_units, n_units)
        - AutoCorr_similarity_matrix.npy: The autocorrelogram similarity matrix of the units (n_units, n_units)
        - PETH_similarity_matrix.npy: The PETH similarity matrix of the units (n_units, n_units)
    """
    data_folder = user_settings["path_to_data"]
    output_folder = user_settings["output_folder"]

    channel_locations = np.load(os.path.join(data_folder, 'channel_locations.npy'))
    isi = np.load(os.path.join(output_folder, 'isi.npy'))
    auto_corr = np.load(os.path.join(output_folder, 'auto_corr.npy'))
    peth = np.load(os.path.join(output_folder, 'peth.npy'))
    n_unit = waveforms.shape[0]

    waveform_similarity_matrix = np.zeros((n_unit, n_unit))
    if 'Waveform' in feature_names:
        waveform_similarity_matrix = computeWaveformSimilarityMatrix(user_settings, waveforms, channel_locations)

    ISI_similarity_matrix = np.zeros((n_unit, n_unit))
    if 'ISI' in feature_names:
        ISI_similarity_matrix = np.corrcoef(isi)
        ISI_similarity_matrix[np.isnan(ISI_similarity_matrix)] = 0
        ISI_similarity_matrix = np.atanh(ISI_similarity_matrix)
        ISI_similarity_matrix = 0.5 * (ISI_similarity_matrix + ISI_similarity_matrix.T) # make it symmetric
        np.fill_diagonal(ISI_similarity_matrix, np.inf)

    AutoCorr_similarity_matrix = np.zeros((n_unit, n_unit))
    if 'AutoCorr' in feature_names:
        AutoCorr_similarity_matrix = np.corrcoef(auto_corr)
        AutoCorr_similarity_matrix[np.isnan(AutoCorr_similarity_matrix)] = 0
        AutoCorr_similarity_matrix = np.atanh(AutoCorr_similarity_matrix)
        AutoCorr_similarity_matrix = 0.5 * (AutoCorr_similarity_matrix + AutoCorr_similarity_matrix.T) # make it symmetric
        np.fill_diagonal(AutoCorr_similarity_matrix, np.inf)

    PETH_similarity_matrix = np.zeros((n_unit, n_unit))
    if 'PETH' in feature_names:
        PETH_similarity_matrix = np.corrcoef(peth)
        PETH_similarity_matrix[np.isnan(PETH_similarity_matrix)] = 0
        PETH_similarity_matrix = np.atanh(PETH_similarity_matrix)
        PETH_similarity_matrix = 0.5 * (PETH_similarity_matrix + PETH_similarity_matrix.T) # make it symmetric
        np.fill_diagonal(PETH_similarity_matrix, np.inf)

    np.save(os.path.join(output_folder, 'waveform_similarity_matrix.npy'), waveform_similarity_matrix)
    np.save(os.path.join(output_folder, 'ISI_similarity_matrix.npy'), ISI_similarity_matrix)
    np.save(os.path.join(output_folder, 'AutoCorr_similarity_matrix.npy'), AutoCorr_similarity_matrix)
    np.save(os.path.join(output_folder, 'PETH_similarity_matrix.npy'), PETH_similarity_matrix)

    feature_names_all = ['Waveform', 'ISI', 'AutoCorr', 'PETH']
    similarity_matrix_all = np.stack((waveform_similarity_matrix, ISI_similarity_matrix, AutoCorr_similarity_matrix, PETH_similarity_matrix), axis=2)

    return (similarity_matrix_all, feature_names_all)

def iterativeClustering(user_settings, similarity_names, waveforms, motion=None):
    """Iterative clustering of the units based on the similarity metrics using HDBSCAN and LDA.
    The similarity metrics are computed firstly, and then HDBSCAN and LDA are performed alternatively to find the best clustering results.
    The clustering results are saved to the output folder.

    Arguments:
        - user_settings (dict): User settings
        - similarity_names (list): The names of the similarity metrics to be computed. The options are 'Waveform', 'ISI', 'AutoCorr', and 'PETH'.

    Outputs:
        - SimilarityMatrix.npy: The similarity matrix of the units
        - SimilarityWeights.npy: The weights of the similarity metrics
        - SimilarityThreshold.npy: The threshold of the similarity metrics from LDA
        - ClusteringResults.npz: The clustering results of the units
        - DistanceMatrix.npy: The distance matrix used for HDBSCAN
        - waveform_similarity_matrix.npy: The waveform similarity matrix of the units
        - ISI_similarity_matrix.npy: The ISI similarity matrix of the units
        - AutoCorr_similarity_matrix.npy: The autocorrelogram similarity matrix of the units
        - PETH_similarity_matrix.npy: The PETH similarity matrix of the units
        - AllSimilarity.npz (optional): The similarity metrics of all units used for clustering
 
    """

    # Load precomputed features
    data_folder = user_settings["path_to_data"]
    output_folder = user_settings["output_folder"]

    sessions = np.load(os.path.join(data_folder , 'session_index.npy'))
    locations = np.load(os.path.join(output_folder, 'locations.npy'))

    # Compute the similarities
    max_distance = user_settings['clustering']['max_distance']
    idx_unit_pairs, _ = getNearbyPairs(max_distance, sessions, locations, motion)
    n_pairs = idx_unit_pairs.shape[0]

    similarity_matrix_all, names_all = computeAllSimilarityMatrix(user_settings, waveforms, similarity_names)
    print("Computing similarity done!")

    n_session = np.max(sessions)

    idx_names = np.array([names_all.index(name) for name in similarity_names])
    similarity_matrix_all = similarity_matrix_all[:, :, idx_names]

    similarity_all = np.zeros((n_pairs, len(similarity_names)))
    for k in range(n_pairs):
        similarity_all[k, :] = similarity_matrix_all[idx_unit_pairs[k, 0], idx_unit_pairs[k, 1], :]

    weights = np.ones(len(similarity_names))/len(similarity_names)
    similarity_matrix = np.sum(similarity_matrix_all*weights, axis=2)

    for iter in range(1, user_settings['clustering']['n_iter']+1):
        print(f'Iteration {iter} starts!')

        # HDBSCAN
        distance_matrix = 1./(1 + np.tanh(similarity_matrix))
        np.fill_diagonal(distance_matrix, 0)
        
        clusterer = hdbscan.HDBSCAN(
            min_samples=1,
            cluster_selection_epsilon=0,
            min_cluster_size=2,
            max_cluster_size=n_session,
            metric='precomputed'
        )
        
        idx_cluster_hdbscan = clusterer.fit_predict(distance_matrix)
        idx_cluster_hdbscan[idx_cluster_hdbscan >= 0] += 1  # MATLAB starts from 1
        
        n_cluster = np.max(idx_cluster_hdbscan)
        hdbscan_matrix = np.zeros_like(similarity_matrix, dtype=bool)
        
        for k in range(1, n_cluster+1):
            idx = np.where(idx_cluster_hdbscan == k)[0]
            for i in range(len(idx)):
                for j in range(i+1, len(idx)):
                    hdbscan_matrix[idx[i], idx[j]] = True
                    hdbscan_matrix[idx[j], idx[i]] = True
        
        np.fill_diagonal(hdbscan_matrix, True)
        
        is_matched = np.array([hdbscan_matrix[idx_unit_pairs[k,0], idx_unit_pairs[k,1]] 
                                for k in range(n_pairs)])
        
        if iter != user_settings['motionEstimation']['n_iter'] - 1:
            # LDA and update weights
            mdl = LinearDiscriminantAnalysis()
            mdl.fit(similarity_all, is_matched)
            
            temp = mdl.coef_[0]
            weights = temp / np.sum(temp)
            print('Weights:')
            print('   '.join(similarity_names))
            print(weights)
            
            # Update the similarity matrix
            similarity_matrix = np.sum(similarity_matrix_all*weights, axis=2)

    Z = clusterer.single_linkage_tree_.to_numpy()
    np.save(os.path.join(output_folder, 'DistanceMatrix.npy'), distance_matrix)
    Z_ordered = optimal_leaf_ordering(Z, squareform(distance_matrix))
    leafOrder = leaves_list(Z_ordered)

    # set the threshold based on LDA results
    thres = mdl.intercept_[0] / (-mdl.coef_[0][0]) * weights[0]

    good_matches_matrix = similarity_matrix > thres
    np.fill_diagonal(good_matches_matrix, True)

    # plot the distribution of similarity
    n_plots = len(similarity_names)
    fig, axes = plt.subplots(1, n_plots, figsize=(5*n_plots, 5))
    for k in range(n_plots):
        axes[k].hist(similarity_all[:, k], bins=50, color='blue', density=True)
        axes[k].set_title(similarity_names[k])
        axes[k].set_xlabel(similarity_names[k] + ' Similarity')
        axes[k].set_ylabel('Density')

    plt.savefig(os.path.join(output_folder, 'Figures/AllSimilarity.png'))
    plt.close()

    # Save the results
    np.save(os.path.join(output_folder, 'SimilarityMatrix.npy'), similarity_matrix)
    np.save(os.path.join(output_folder, 'SimilarityWeights.npy'), weights)
    np.save(os.path.join(output_folder, 'SimilarityThreshold.npy'), thres)
    np.save(os.path.join(output_folder, 'SimilarityPairs.npy'), idx_unit_pairs)
    np.save(os.path.join(output_folder, 'ClusterMatrix.npy'), hdbscan_matrix)

    np.savez(os.path.join(user_settings['output_folder'], 'ClusteringResults.npz'),
        weights=weights, 
        similarity_all=similarity_all, idx_unit_pairs=idx_unit_pairs,
        thres=thres, good_matches_matrix=good_matches_matrix,
        similarity_matrix=similarity_matrix, distance_matrix=distance_matrix,
        leafOrder=leafOrder,
        idx_cluster_hdbscan=idx_cluster_hdbscan, hdbscan_matrix=hdbscan_matrix,
        n_cluster=n_cluster)



import numpy as np
from joblib import Parallel, delayed
from tqdm import tqdm
import os
from .utils import waveformEstimation
import copy

def computeWaveformFeatures(user_settings, waveform_all, motion):
    """ Compute the corrected waveforms based on the motion of the probe.
    The corrected waveforms on the reference probe are computed using the Kriging interpolation method
    and saved to the output folder.

    Arguments:
        - user_settings (dict): User settings
        - waveform_all (numpy.ndarray): The waveforms of all units (n_unit, n_channel, n_sample)
        - motion (Motion): The motion object containing the linear and constant parameters for correction
    Outputs:
        - waveforms_corrected.npy: The corrected waveforms.
    
    """

    data_folder = user_settings["path_to_data"]
    output_folder = user_settings["output_folder"]
    n_templates = user_settings["waveformCorrection"]["n_templates"]

    channel_locations = np.load(os.path.join(data_folder, 'channel_locations.npy'))
    sessions = np.load(os.path.join(data_folder , 'session_index.npy'))

    locations = np.load(os.path.join(output_folder, 'locations.npy'))

    n_sample = waveform_all.shape[2]
    n_channel = waveform_all.shape[1]
    n_unit = waveform_all.shape[0]

    min_channel_depth = np.min(channel_locations[:,1])
    max_channel_depth = np.max(channel_locations[:,1])

    motion_bottom = motion.LinearScale*motion.Linear*min_channel_depth + motion.Constant
    motion_top = motion.LinearScale*motion.Linear*max_channel_depth + motion.Constant

    min_motion = np.min(np.concatenate((motion_bottom, motion_top)))
    max_motion = np.max(np.concatenate((motion_bottom, motion_top)))
    print('The range of motion: [%.1f μm ~ %.1f μm]\n' % (min_motion, max_motion))

    def process_spike(locations_this, motion, channel_locations, waveform_this, session_this, n_templates, min_motion, max_motion):

        motion_this = copy.deepcopy(motion)
        waveforms_corrected = np.zeros((n_channel, n_sample, n_templates))

        for k in range(n_templates):
            if n_templates == 2:
                if k == 0:
                    motion_this.Constant = motion_this.Constant - min_motion
                else:
                    motion_this.Constant = motion_this.Constant - max_motion

            dy = motion_this.get_motion(session_this, locations_this[1])
            location_new = locations_this.copy()
            location_new[1] -= dy

            waveforms_corrected[:,:,k] = waveformEstimation(
                waveform_this, locations_this, channel_locations, location_new)
            
        return waveforms_corrected

    # Run parallel processing with progress bar
    out = Parallel(n_jobs=user_settings["n_jobs"])(
        delayed(process_spike)(locations[k,:2], motion, channel_locations, waveform_all[k,:,:], sessions[k], n_templates, min_motion, max_motion) 
        for k in tqdm(range(n_unit), desc='Computing waveform features')
    )

    waveforms_corrected = np.zeros((n_unit, n_channel, n_sample, n_templates))
    for k in range(n_unit):
        waveforms_corrected[k,:,:,:] = out[k]

    # Save the corrected waveforms
    output_folder = user_settings['output_folder']
    np.save(os.path.join(output_folder, 'waveforms_corrected.npy'), waveforms_corrected)



import numpy as np
from scipy.optimize import least_squares
from scipy.sparse import csgraph
import os

def computeKernel2D(xp, yp, sig=20):
    '''Compute the 2D kernel matrix for the given points xp and yp.

    Arguments:
        - xp: 2D array of points (n_samples, 2)
        - yp: 2D array of points (n_samples, 2)
        - sig: standard deviation for the Gaussian kernel (default: 20)

    Returns:
        - K: kernel matrix (n_samples_xp, n_samples_yp)
    '''

    distx = np.abs(
        np.expand_dims(xp[:,0], 1) - np.expand_dims(yp[:,0], 0))
    disty = np.abs(
        np.expand_dims(xp[:,1], 1) - np.expand_dims(yp[:,1], 0))

    sigx = sig
    sigy = 1.5 * sig

    p = 1
    K = np.exp(-(distx / sigx)**p - (disty / sigy)**p)
    return K

def graphEditNumber(matA, matB):
    ''' Compute the merge number of two graphs A and B and the number of same merges.

    Arguments:
        - matA: connectivity matrix of graph A (n_nodes_A, n_nodes_A)
        - matB: connectivity matrix of graph B (n_nodes_B, n_nodes_B)

    Returns:
        - nSame: number of same merges
        - nA: number of merges in graph A
        - nB: number of merges in graph B
    '''

    # Convert to sparse matrices and find connected components
    comp_A = csgraph.connected_components(matA, directed=False)[1]
    comp_B = csgraph.connected_components(matB, directed=False)[1]
    comp_AB = csgraph.connected_components(matA*matB, directed=False)[1]
    
    # Count edges within components (n-1 edges per n-node component)
    nA = sum([np.sum(comp_A == i)-1 for i in np.unique(comp_A)])
    nB = sum([np.sum(comp_B == i)-1 for i in np.unique(comp_B)])
    nSame = sum([np.sum(comp_AB == i)-1 for i in np.unique(comp_AB)])
    return (nSame, nA, nB)

def spikeLocation(waveforms_mean, channel_locations, n_nearest_channels=20, algorithm='monopolar_triangulation'):
    '''Spike location estimation using either center_of_mass or monopolar_triangulation
    
    monopolar_triangulation: refer to Boussard, Julien, Erdem Varol, Hyun Dong Lee, Nishchal Dethe, and Liam Paninski. “Three-Dimensional Spike Localization and Improved Motion Correction for Neuropixels Recordings.” In Advances in Neural Information Processing Systems, 34:22095–105. Curran Associates, Inc., 2021. https://proceedings.neurips.cc/paper/2021/hash/b950ea26ca12daae142bd74dba4427c8-Abstract.html.
    > https://spikeinterface.readthedocs.io/en/stable/modules/postprocessing.html#spike-locations
    > https://github.com/SpikeInterface/spikeinterface/blob/main/src/spikeinterface/postprocessing/localization_tools.py#L334

    Arguments:
        - waveforms_mean: mean waveforms (n_channels, n_samples)
        - channel_locations: 2D array of channel locations (n_channels, 2)
        - n_nearest_channels: number of nearest channels to consider for localization, default is 20
        - algorithm: 'center_of_mass' or 'monopolar_triangulation', default is 'monopolar_triangulation'

    returns:
        - x: x coordinate of the spike location
        - y: y coordinate of the spike location
        - z: z coordinate of the spike location
        - ptt: peak-to-trough value of the spike waveform

    '''

    # get n_nearest_channels from the channels with the largest peak-to-trough value
    peaks_to_trough = np.max(waveforms_mean, axis=1) - np.min(waveforms_mean, axis=1)
    idx_max = np.argmax(peaks_to_trough)

    loc_max = channel_locations[idx_max, :]
    distance_to_max = np.sum((channel_locations - loc_max)**2, axis=1)

    idx_sorted = np.argsort(distance_to_max)
    idx_included = idx_sorted[:n_nearest_channels]

    # calculate the center_to_mass location
    ptt_max = peaks_to_trough[idx_max]
    ptt_this = peaks_to_trough[idx_included]
    loc_this = channel_locations[idx_included,:]

    loc_center_to_mass = np.sum(loc_this * ptt_this[:, np.newaxis], axis=0) / np.sum(ptt_this)

    if algorithm.lower() == 'center_of_mass':
        x = loc_center_to_mass[0]
        y = loc_center_to_mass[1]
        z = 0
        ptt = ptt_max

        return x, y, z, ptt

    # calculate the monopolar_triangulation location
    def fun(x, ptt, loc_this):
        ptt_estimated = x[3] / np.sqrt((loc_this[:,0]-x[0])**2 + (loc_this[:,1]-x[1])**2 + x[2]**2)
        return ptt - ptt_estimated

    x0 = [loc_center_to_mass[0], loc_center_to_mass[1], 1, ptt_max]
    bounds = (
        [x0[0] - 100, x0[1] - 100, 1, 0],
        [x0[0] + 100, x0[1] + 100, 100 * 10, 1000*ptt_max],
    )

    output = least_squares(fun, x0=x0, bounds=bounds, args=(ptt_this, loc_this))
    
    return tuple(output["x"])

def waveformEstimation(waveform_mean, location, channel_locations, location_new):
    '''Waveform estimation with Kriging interpolation.

    Arguments:
        - waveform_mean: mean waveform (n_channels, n_samples)
        - location: original location of the spike (x, y), 1D array of length 2
        - channel_locations: 2D array of channel locations (n_channels, 2)
        - location_new: new location of the spike (x, y), 1D array of length 2
    
    Returns:
        - waveform_out: estimated waveform at the new location (n_samples)
    '''
    
    # Calculate mapped location
    location_mapped_to_old = channel_locations - (np.array(location_new) - np.array(location))
    
    # 2D coordinates for interpolation
    xp = channel_locations
    
    # 2D kernel of the original channel positions
    Kxx = computeKernel2D(xp, xp)
    
    # 2D kernel of the new channel positions
    yp = location_mapped_to_old
    Kyx = computeKernel2D(yp, xp)
    
    # Kernel prediction matrix
    M = Kyx @ np.linalg.inv(Kxx + 0.01 * np.eye(Kxx.shape[0]))
    
    # Interpolated waveform at the new location
    waveform_out = M @ waveform_mean
    
    return waveform_out

def corrcoef2(x, y):
    '''Compute the Pearson correlation coefficient between two matrices x and y.

    Arguments:
        - x: 2D array of shape (n_samples, n_features_x)
        - y: 2D array of shape (n_samples, n_features_y)

    Returns:
        - r: Pearson correlation coefficient matrix of shape (n_features_x, n_features_y)
    '''

    x = (x - x.mean(axis=0))/x.std(axis=0)
    y = (y - y.mean(axis=0))/y.std(axis=0)

    r = np.dot(x.T, y) / x.shape[0]
    return r

def computeAutoCorr(spike_times, window, binwidth):
    '''Compute the autocorrelation of spike times.
    
    Refer to the elegant Python impletantation from phylib:
        https://github.com/cortex-lab/phylib/blob/master/phylib/stats/ccg.py#L34

    Arguments:
        - spike_times: 1D array of spike times
        - window: time window for autocorrelation (in ms, default: 300 ms)
        - binwidth: width of the bins for histogram (in ms, default: 1 ms)

    Returns:
        - auto_corr: autocorrelation values
        - lag: lag values
    '''

    n_bins = np.int64(np.floor(window/binwidth)+1)
    auto_corr_right = np.zeros(n_bins)  # the right side of auto_corr

    shift = 1
    while True:
        dt = spike_times[shift:] - spike_times[:-shift]
        i_bin = np.int64(np.round(dt / binwidth))
        i_bin = i_bin[i_bin < n_bins]

        if i_bin.size == 0:
            break

        counts = np.bincount(i_bin)

        auto_corr_right[:len(counts)] += counts

        shift += 1

    auto_corr = np.concatenate((np.flip(auto_corr_right[1:]), auto_corr_right))
    lag = np.arange(-(n_bins-1)*binwidth, (n_bins)*binwidth, binwidth)

    assert(len(auto_corr) == len(lag))
    assert(np.sum(lag==0) > 0)

    return auto_corr, lag


class Motion:
    """Class to handle motion estimation data.
    This class allows for saving and loading motion data, as well as retrieving motion values for specific sessions.

    Attributes:
        - LinearScale: scaling factor for linear motion (default: 0.001)
        - Linear: linear motion parameters for each session (if num_sessions is provided)
        - Constant: constant motion parameters for each session (if num_sessions is provided)

    Methods:
        - __init__(num_sessions=None): Initializes the Motion object.
        - save(output_folder): Saves the motion data to a file.
        - load(output_folder): Loads the motion data from a file.
        - get_motion(session, depth=None): Retrieves the motion for a specific session, optionally considering depth.  

    """

    def __init__(self, num_sessions=None):
        """Initialize the Motion class.
        If num_sessions is None, Linear and Constant will be set to None.
        Otherwise, they will be initialized as zero arrays of length num_sessions.

        Args:
            num_sessions (int, optional): Number of sessions. If None, Linear and Constant will be None.

        Returns:
            None

        """
        self.LinearScale = np.array(0.001, dtype=np.float64)

        if num_sessions is None:
            self.Linear = None
            self.Constant = None
        else:
            self.Linear = np.zeros(num_sessions, dtype=np.float64)
            self.Constant = np.zeros(num_sessions, dtype=np.float64)

    def save(self, output_folder):
        """Save the motion data to a file.

        Args:
            output_folder (str): Path to the folder where the motion data will be saved.

        Returns:
            None

        """
        np.save(os.path.join(output_folder, 'motion_linear_scale.npy'), self.LinearScale)
        np.save(os.path.join(output_folder, 'motion_linear.npy'), self.Linear)
        np.save(os.path.join(output_folder, 'motion_constant.npy'), self.Constant)

    @staticmethod
    def load(output_folder):
        """Load the motion data from a file.

        Args:
            output_folder (str): Path to the folder where the motion data is saved.

        Returns:
            Motion: An instance of the Motion class with loaded data.

        """
        motion = Motion()
        motion.LinearScale = np.load(os.path.join(output_folder, 'motion_linear_scale.npy'))
        motion.Linear = np.load(os.path.join(output_folder, 'motion_linear.npy'))
        motion.Constant = np.load(os.path.join(output_folder, 'motion_constant.npy'))
        return motion

    def get_motion(self, session, depth=None):
        """Get the motion for a specific session.

        Args:
            session (int): The session number.
            depth (float, optional): The depth value. If None, only the constant motion is returned.

        Returns:
            float: The motion value for the specified session and depth.

        """
        if depth is None:
            return self.Constant[session - 1]
        
        return self.LinearScale * self.Linear[session - 1] * depth + self.Constant[session - 1]



from .Preprocess import preprocess, spikeInfo2npy
from .MotionEstimation import motionEstimation, computeMotion, Motion
from .ComputeWaveformFeatures import computeWaveformFeatures
from .IterativeClustering import finalClustering
from .AutoCuration import autoCuration
